/*
 * Copyright (c) 2002-2003 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 *
 * Formatting notes:
 * This code follows the "Whitesmiths style" C indentation rules. Plenty of discussion
 * on C indentation can be found on the web, such as <http://www.kafejo.com/komp/1tbs.htm>,
 * but for the sake of brevity here I will say just this: Curly braces are not syntactially
 * part of an "if" statement; they are the beginning and ending markers of a compound statement;
 * therefore common sense dictates that if they are part of a compound statement then they
 * should be indented to the same level as everything else in that compound statement.
 * Indenting curly braces at the same level as the "if" implies that curly braces are
 * part of the "if", which is false. (This is as misleading as people who write "char* x,y;"
 * thinking that variables x and y are both of type "char*" -- and anyone who doesn't
 * understand why variable y is not of type "char*" just proves the point that poor code
 * layout leads people to unfortunate misunderstandings about how the C language really works.)

	Change History (most recent first):

$Log: mDNSPosix.c,v $
Revision 1.25.2.1  2004/04/09 17:57:31  cheshire
Make sure to set the TxAndRx field so that duplicate suppression works correctly

Revision 1.25  2003/10/30 19:25:49  cheshire
Fix signed/unsigned warning on certain compilers

Revision 1.24  2003/08/18 23:12:23  cheshire
<rdar://problem/3382647> mDNSResponder divide by zero in mDNSPlatformTimeNow()

Revision 1.23  2003/08/12 19:56:26  cheshire
Update to APSL 2.0

Revision 1.22  2003/08/06 18:46:15  cheshire
LogMsg() errors are serious -- always report them to stderr, regardless of debugging level

Revision 1.21  2003/08/06 18:20:51  cheshire
Makefile cleanup

Revision 1.20  2003/08/05 23:56:26  cheshire
Update code to compile with the new mDNSCoreReceive() function that requires a TTL
(Right now mDNSPosix.c just reports 255 -- we should fix this)

Revision 1.19  2003/07/19 03:15:16  cheshire
Add generic MemAllocate/MemFree prototypes to mDNSPlatformFunctions.h,
and add the obvious trivial implementations to each platform support layer

Revision 1.18  2003/07/14 18:11:54  cheshire
Fix stricter compiler warnings

Revision 1.17  2003/07/13 01:08:38  cheshire
There's not much point running mDNS over a point-to-point link; exclude those

Revision 1.16  2003/07/02 21:19:59  cheshire
<rdar://problem/3313413> Update copyright notices, etc., in source code comments

Revision 1.15  2003/06/18 05:48:41  cheshire
Fix warnings

Revision 1.14  2003/05/26 03:21:30  cheshire
Tidy up address structure naming:
mDNSIPAddr         => mDNSv4Addr (for consistency with mDNSv6Addr)
mDNSAddr.addr.ipv4 => mDNSAddr.ip.v4
mDNSAddr.addr.ipv6 => mDNSAddr.ip.v6

Revision 1.13  2003/05/26 03:01:28  cheshire
<rdar://problem/3268904> sprintf/vsprintf-style functions are unsafe; use snprintf/vsnprintf instead

Revision 1.12  2003/05/21 03:49:18  cheshire
Fix warning

Revision 1.11  2003/05/06 00:00:50  cheshire
<rdar://problem/3248914> Rationalize naming of domainname manipulation functions

Revision 1.10  2003/04/25 01:45:57  cheshire
<rdar://problem/3240002> mDNS_RegisterNoSuchService needs to include a host name

Revision 1.9  2003/03/20 21:10:31  cheshire
Fixes done at IETF 56 to make mDNSProxyResponderPosix run on Solaris

Revision 1.8  2003/03/15 04:40:38  cheshire
Change type called "mDNSOpaqueID" to the more descriptive name "mDNSInterfaceID"

Revision 1.7  2003/03/13 03:46:21  cheshire
Fixes to make the code build on Linux

Revision 1.6  2003/03/08 00:35:56  cheshire
Switched to using new "mDNS_Execute" model (see "mDNSCore/Implementer Notes.txt")

Revision 1.5  2002/12/23 22:13:31  jgraessl
Reviewed by: Stuart Cheshire
Initial IPv6 support for mDNSResponder.

Revision 1.4  2002/09/27 01:47:45  cheshire
Workaround for Linux 2.0 systems that don't have IP_PKTINFO

Revision 1.3  2002/09/21 20:44:53  zarzycki
Added APSL info

Revision 1.2  2002/09/19 21:25:36  cheshire
mDNS_snprintf() doesn't need to be in a separate file

Revision 1.1  2002/09/17 06:24:34  cheshire
First checkin
*/

#if (OEM_TARGET_OS == VXWORKS)
#include "nest.h"
#include "compoem.h"
#endif /* (OEM_TARGET_OS == VXWORKS) */

#if MDNS_OEM_COMPONENT_INCLUDED

#if (OEM_TARGET_OS == VXWORKS)
#include "types/vxTypesBase.h"
#include "semLib.h"
#include "sockLib.h"
#include "tickLib.h"
#include "timers.h"
#include "ioLib.h"
#include "poseapi.h"
#include "etherLib.h"
#include "hostLib.h"
#include "inetLib.h"
#include "vx_iputils.h"
#endif /* (OEM_TARGET_OS == VXWORKS) */

#include "mDNSClientAPI.h"           // Defines the interface provided to the client layer above
#include "mDNSPlatformFunctions.h"   // Defines the interface to the supporting layer below
#if (OEM_TARGET_OS != VXWORKS)
#include "mDNSPosix.h"				 // Defines the specific types needed to run mDNS on this platform
#endif /* OEM_TARGET_OS != VXWORKS */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#if (OEM_TARGET_OS != VXWORKS)
#include <sys/uio.h>
#endif /* (OEM_TARGET_OS != VXWORKS) */
#if (OEM_TARGET_OS == VXWORKS)
#include "time.h"
static char *strdup(const char *s) {
char *temp=malloc(strlen(s)+1);
if (temp==NULL) return(NULL);
else return(strcpy(temp,s));
}
#include "NetRegionImports.h"
#include "ctype.h"
#include "compoem.h"
#include "stdbool.h"
#include "xpm/xpmapi.h"
#include <opb_eventlog.h>
#include "osXpmDefines.h"
#include "sys/netXpmDefines.h"
#endif /* (OEM_TARGET_OS == VXWORKS) */
#include <netinet/in.h>

#if (OEM_TARGET_OS != VXWORKS)
#include "mDNSUNP.h"
#else
#include "mDNS_UNP.h"
#include "mDNSPlatformPosix.h"	     // Defines the specific types needed to run mDNS on this platform
#include "mDNSPlatformEnvironment.h"    // Defines the specific types needed to run mDNS on this platform

#include "pipeDrv.h"
#define	kMDNSPipeName					"/pipe/mDNS"
#define	kMDNSPipeMessageQueueSize		16
#define	kMDNSPipeMessageSize			1

typedef uint8_t		MDNSPipeCommandCode;
enum
{
	kMDNSPipeCommandCodeInvalid			= 0, 
	kMDNSPipeCommandCodeTimer 			= 1, 
	kMDNSPipeCommandCodeUpdateIP		= 2, 
	kMDNSPipeCommandCodeQuit			= 3, 
	kMDNSPipeCommandCodeNameCollision	= 4,
	kMDNSPipeCommandCodeIFLinkChange	= 5
};

#endif /* (OEM_TARGET_OS != VXWORKS) */

// ***************************************************************************
// Structures

// PosixNetworkInterface is a record extension of the core NetworkInterfaceInfo
// type that supports extra fields needed by the Posix platform.
//
// IMPORTANT: coreIntf must be the first field in the structure because
// we cast between pointers to the two different types regularly.

typedef struct PosixNetworkInterface PosixNetworkInterface;

struct PosixNetworkInterface
	{
	NetworkInterfaceInfo    coreIntf;
	const char *            intfName;
	PosixNetworkInterface * aliasIntf;
	int                     index;
	int                     multicastSocket;
	int                     multicastSocketv6;
	};

// ***************************************************************************
// Globals (for debugging)

static int num_registered_interfaces = 0;
static int num_pkts_accepted = 0;
static int num_pkts_rejected = 0;

// ***************************************************************************
// Functions

int gMDNSPlatformPosixVerboseLevel = 0;

// Note, this uses mDNS_vsnprintf instead of standard "vsnprintf", because mDNS_vsnprintf knows
// how to print special data types like IP addresses and length-prefixed domain names
mDNSexport void debugf_(const char *format, ...)
	{
	unsigned char buffer[512];
	va_list ptr;
	va_start(ptr,format);
	buffer[mDNS_vsnprintf((char *)buffer, sizeof(buffer), format, ptr)] = 0;
	va_end(ptr);
	if (gMDNSPlatformPosixVerboseLevel >= 1)
		fprintf(stderr, "%s\n", buffer);
	fflush(stderr);
	}

mDNSexport void verbosedebugf_(const char *format, ...)
	{
	unsigned char buffer[512];
	va_list ptr;
	va_start(ptr,format);
	buffer[mDNS_vsnprintf((char *)buffer, sizeof(buffer), format, ptr)] = 0;
	va_end(ptr);
	if (gMDNSPlatformPosixVerboseLevel >= 2)
		fprintf(stderr, "%s\n", buffer);
	fflush(stderr);
	}

mDNSexport void LogMsg(const char *format, ...)
	{
	unsigned char buffer[512];
	va_list ptr;
	va_start(ptr,format);
	buffer[mDNS_vsnprintf((char *)buffer, sizeof(buffer), format, ptr)] = 0;
	va_end(ptr);
	fprintf(stderr, "%s\n", buffer);
	fflush(stderr);
	}

#define PosixErrorToStatus(errNum) ((errNum) == 0 ? mStatus_NoError : mStatus_UnknownErr)

static void SockAddrTomDNSAddr(const struct sockaddr *const sa, mDNSAddr *ipAddr, mDNSIPPort *ipPort)
	{
	switch (sa->sa_family)
		{
		case AF_INET:
			{
			struct sockaddr_in* sin          = (struct sockaddr_in*)sa;
			ipAddr->type                     = mDNSAddrType_IPv4;
			ipAddr->ip.v4.NotAnInteger       = sin->sin_addr.s_addr;
			if (ipPort) ipPort->NotAnInteger = sin->sin_port;
			break;
			}

#ifdef mDNSIPv6Support
		case AF_INET6:
			{
			struct sockaddr_in6* sin6        = (struct sockaddr_in6*)sa;
			assert(sin6->sin6_len == sizeof(*sin6));
			ipAddr->type                     = mDNSAddrType_IPv6;
			ipAddr->ip.v6                    = *(mDNSv6Addr*)&sin6->sin6_addr;
			if (ipPort) ipPort->NotAnInteger = sin6->sin6_port;
			break;
			}
#endif

		default:
			verbosedebugf("SockAddrTomDNSAddr: Uknown address family %d\n", sa->sa_family);
			ipAddr->type = mDNSAddrType_None;
			if (ipPort) ipPort->NotAnInteger = 0;
			break;
		}
	}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Send and Receive
#endif

// mDNS core calls this routine when it needs to send a packet.
mDNSexport mStatus mDNSPlatformSendUDP(const mDNS *const m, const DNSMessage *const msg, const mDNSu8 *const end,
	mDNSInterfaceID InterfaceID, mDNSIPPort srcPort, const mDNSAddr *dst, mDNSIPPort dstPort)
	{
	int                     err;
	struct sockaddr_storage to;
	PosixNetworkInterface * thisIntf;

	assert(m != NULL);
	assert(msg != NULL);
	assert(end != NULL);
	assert( (((char *) end) - ((char *) msg)) > 0 );
	assert(InterfaceID != 0); // Can't send from zero source address
	assert(srcPort.NotAnInteger != 0);     // Nor from a zero source port
	assert(dstPort.NotAnInteger != 0);     // Nor from a zero source port

	if (dst->type == mDNSAddrType_IPv4)
		{
		struct sockaddr_in *sin = (struct sockaddr_in*)&to;
#ifndef NOT_HAVE_SA_LEN
		sin->sin_len            = sizeof(*sin);
#endif
		sin->sin_family         = AF_INET;
		sin->sin_port           = dstPort.NotAnInteger;
		sin->sin_addr.s_addr    = dst->ip.v4.NotAnInteger;
		}

#ifdef mDNSIPv6Support
	else if (dst->type == mDNSAddrType_IPv6)
		{
		struct sockaddr_in6 *sin6 = (struct sockaddr_in6*)&to;
		mDNSPlatformMemZero(sin6, sizeof(*sin6));
		sin6->sin6_len            = sizeof(*sin6);
		sin6->sin6_family         = AF_INET6;
		sin6->sin6_port           = dstPort.NotAnInteger;
		sin6->sin6_addr           = *(struct in6_addr*)&dst->ip.v6;
		}
#endif

	err = 0;
	thisIntf = (PosixNetworkInterface *)(InterfaceID);
	if (dst->type == mDNSAddrType_IPv4)
#if (OEM_TARGET_OS != VXWORKS)
		err = sendto(thisIntf->multicastSocket, msg, (char*)end - (char*)msg, 0, (struct sockaddr *)&to, GET_SA_LEN(to));
#else
		err = sendto(thisIntf->multicastSocket, (char *)msg, (char*)end - (char*)msg, 0, (struct sockaddr *)&to, GET_SA_LEN(to));
#endif /* (OEM_TARGET_OS != VXWORKS) */

#ifdef mDNSIPv6Support
	else if (dst->type == mDNSAddrType_IPv6)
		err = sendto(thisIntf->multicastSocketv6, msg, (char*)end - (char*)msg, 0, (struct sockaddr *)&to, GET_SA_LEN(to));
#endif

	if (err > 0) err = 0;
	else if (err < 0)
		verbosedebugf("mDNSPlatformSendUDP got error %d (%s) sending packet to %#a on interface %#a/%s/%d",
					  errno, strerror(errno), dst, &thisIntf->coreIntf.ip, thisIntf->intfName, thisIntf->index);

	return PosixErrorToStatus(err);
	}

// This routine is called when the main loop detects that data is available on a socket.
static void SocketDataReady(mDNS *const m, PosixNetworkInterface *intf, int skt)
	{
	mDNSAddr   senderAddr, destAddr;
	mDNSIPPort senderPort;
	ssize_t                 packetLen;
	DNSMessage              packet;
	struct my_in_pktinfo    packetInfo;
	struct sockaddr_storage from;
	socklen_t               fromLen;
	int                     flags;
	mDNSBool                reject;

	assert(m    != NULL);
	assert(intf != NULL);
	assert(skt  >= 0);

	fromLen = sizeof(from);
	flags   = 0;
	packetLen = recvfrom_flags(skt, &packet, sizeof(packet), &flags, (struct sockaddr *) &from, &fromLen, &packetInfo);

	if (packetLen >= 0)
		{
		SockAddrTomDNSAddr((struct sockaddr*)&from, &senderAddr, &senderPort);
		SockAddrTomDNSAddr((struct sockaddr*)&packetInfo.ipi_addr, &destAddr, NULL);
		// If we have broken IP_RECVDSTADDR functionality (so far
		// I've only seen this on OpenBSD) then apply a hack to
		// convince mDNS Core that this isn't a spoof packet.
		// Basically what we do is check to see whether the
		// packet arrived as a multicast and, if so, set its
		// destAddr to the mDNS address.
		//
		// I must admit that I could just be doing something
		// wrong on OpenBSD and hence triggering this problem
		// but I'm at a loss as to how.
		//
		// If this platform doesn't have IP_PKTINFO or IP_RECVDSTADDR, then we have
		// no way to tell the destination address or interface this packet arrived on,
		// so all we can do is just assume it's a multicast

#if (OEM_TARGET_OS == VXWORKS)
		// vxWorks does not fill packetInfo due to a problem
		// in udp_usrreq.c handling multi-/broad-cast packets
#define HAVE_BROKEN_RECVDSTADDR 1
#endif /* (OEM_TARGET_OS == VXWORKS) */
		#if HAVE_BROKEN_RECVDSTADDR || (!defined(IP_PKTINFO) && !defined(IP_RECVDSTADDR))
#if (OEM_TARGET_OS != VXWORKS)
			if ( (destAddr.NotAnInteger == 0) && (flags & MSG_MCAST) )
				{
				destAddr.type == senderAddr.type;
				if      (senderAddr.type == mDNSAddrType_IPv4) destAddr.ip.v4 = AllDNSLinkGroup;
				else if (senderAddr.type == mDNSAddrType_IPv6) destAddr.ip.v6 = AllDNSLinkGroupv6;
				}
#else
			if (destAddr.type == mDNSAddrType_None)
				destAddr.type = senderAddr.type;
			//destAddr may be have invalid value from recvfrom_flags
				if      (senderAddr.type == mDNSAddrType_IPv4) destAddr.ip.v4 = AllDNSLinkGroup;
				else if (senderAddr.type == mDNSAddrType_IPv6) destAddr.ip.v6 = AllDNSLinkGroupv6;
#endif /* (OEM_TARGET_OS != VXWORKS) */
		#endif

		// We only accept the packet if the interface on which it came
		// in matches the interface associated with this socket.
		// We do this match by name or by index, depending on which
		// information is available.  recvfrom_flags sets the name
		// to "" if the name isn't available, or the index to -1
		// if the index is available.  This accomodates the various
		// different capabilities of our target platforms.

		reject = mDNSfalse;
		if      ( packetInfo.ipi_ifname[0] != 0 ) reject = (strcmp(packetInfo.ipi_ifname, intf->intfName) != 0);
		else if ( packetInfo.ipi_ifindex != -1 )  reject = (packetInfo.ipi_ifindex != intf->index);

		if (reject)
			{
			verbosedebugf("SocketDataReady ignored a packet from %#a to %#a on interface %s/%d expecting %#a/%s/%d",
				&senderAddr, &destAddr, packetInfo.ipi_ifname, packetInfo.ipi_ifindex,
				&intf->coreIntf.ip, intf->intfName, intf->index);
			packetLen = -1;
			num_pkts_rejected++;
			if (num_pkts_rejected > (num_pkts_accepted + 1) * (num_registered_interfaces + 1) * 2)
				{
				fprintf(stderr,
					"*** WARNING: Received %d packets; Accepted %d packets; Rejected %d packets because of interface mismatch\n",
					num_pkts_accepted + num_pkts_rejected, num_pkts_accepted, num_pkts_rejected);
				num_pkts_accepted = 0;
				num_pkts_rejected = 0;
				}
			}
		else
			{
			verbosedebugf("SocketDataReady got a packet from %#a to %#a on interface %#a/%s/%d",
				&senderAddr, &destAddr, &intf->coreIntf.ip, intf->intfName, intf->index);
			num_pkts_accepted++;
			}
		}

	if (packetLen >= 0 && packetLen < (ssize_t)sizeof(DNSMessageHeader))
		{
		debugf("SocketDataReady packet length (%d) too short", packetLen);
		packetLen = -1;
		}

	if (packetLen >= 0)
		mDNSCoreReceive(m, &packet, (mDNSu8 *)&packet + packetLen,
			&senderAddr, senderPort, &destAddr, MulticastDNSPort, intf->coreIntf.InterfaceID, 255);
	}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Init and Term
#endif

// On OS X this gets the text of the field labelled "Computer Name" in the Sharing Prefs Control Panel
// Other platforms can either get the information from the appropriate place,
// or they can alternatively just require all registering services to provide an explicit name
#if (OEM_TARGET_OS != VXWORKS)
mDNSlocal void GetUserSpecifiedFriendlyComputerName(domainlabel *const namelabel)
#else
mDNSlocal void GetUserSpecifiedFriendlyComputerName(mDNS *const m, domainlabel *const namelabel)
#endif /* (OEM_TARGET_OS != VXWORKS) */
	{
#if (OEM_TARGET_OS != VXWORKS)
	MakeDomainLabelFromLiteralString(namelabel, "Fill in Default Service Name Here");
#else
	MakeDomainLabelFromLiteralString(namelabel, m->p->trackedNCLvars.perDeviceServiceRoot);
#endif /* (OEM_TARGET_OS != VXWORKS) */
	}

// This gets the current hostname, truncating it at the first dot if necessary
#if (OEM_TARGET_OS != VXWORKS)
mDNSlocal void GetUserSpecifiedRFC1034ComputerName(domainlabel *const namelabel)
#else
mDNSlocal void GetUserSpecifiedRFC1034ComputerName(mDNS *const m, domainlabel *const namelabel)
#endif /* (OEM_TARGET_OS != VXWORKS) */
	{
	int len = 0;
#if (OEM_TARGET_OS == VXWORKS)
	if (m->p->trackedNCLvars.hostName[0]) {
	    len = strlen(m->p->trackedNCLvars.hostName);
	    if (len >= MAX_DOMAIN_LABEL)
		len = MAX_DOMAIN_LABEL - 1;
	    strncpy(&namelabel->c[1], m->p->trackedNCLvars.hostName, len);
	    namelabel->c[len + 1] = '\0';
	    namelabel->c[0] = len;
	    return;
	}
#endif /* (OEM_TARGET_OS != VXWORKS) */
	gethostname((char *)&namelabel->c[1], MAX_DOMAIN_LABEL);
	while (len < MAX_DOMAIN_LABEL && namelabel->c[len+1] && namelabel->c[len+1] != '.') len++;
	namelabel->c[0] = len;
	}

// Searches the interface list looking for the named interface.
// Returns a pointer to if it found, or NULL otherwise.
static PosixNetworkInterface *SearchForInterfaceByName(mDNS *const m, const char *intfName)
	{
	PosixNetworkInterface *intf;

	assert(m != NULL);
	assert(intfName != NULL);

	intf = (PosixNetworkInterface*)(m->HostInterfaces);
	while ( (intf != NULL) && (strcmp(intf->intfName, intfName) != 0) )
		intf = (PosixNetworkInterface *)(intf->coreIntf.next);

	return intf;
	}

// Frees the specified PosixNetworkInterface structure. The underlying
// interface must have already been deregistered with the mDNS core.
static void FreePosixNetworkInterface(PosixNetworkInterface *intf)
	{
	assert(intf != NULL);
	if (intf->intfName != NULL)        free((void *)intf->intfName);
#if (OEM_TARGET_OS != VXWORKS)
	if (intf->multicastSocket   != -1) assert(close(intf->multicastSocket) == 0);
	if (intf->multicastSocketv6 != -1) assert(close(intf->multicastSocketv6) == 0);
#else
	if (intf->multicastSocket   != -1) {
	    assert(close(intf->multicastSocket) == 0);
	    intf->multicastSocket = -1;  // OPB: minor cleanup
	}
	if (intf->multicastSocketv6 != -1) {
	    assert(close(intf->multicastSocketv6) == 0);
	    intf->multicastSocketv6 = -1;
	}
#endif /* (OEM_TARGET_OS == VXWORKS) */
	free(intf);
	}

// Grab the first interface, deregister it, free it, and repeat until done.
static void ClearInterfaceList(mDNS *const m)
	{
	assert(m != NULL);

	while (m->HostInterfaces)
		{
		PosixNetworkInterface *intf = (PosixNetworkInterface*)(m->HostInterfaces);
		mDNS_DeregisterInterface(m, &intf->coreIntf);
		if (gMDNSPlatformPosixVerboseLevel > 0) fprintf(stderr, "Deregistered interface %s\n", intf->intfName);
		FreePosixNetworkInterface(intf);
		}
	num_registered_interfaces = 0;
	num_pkts_accepted = 0;
	num_pkts_rejected = 0;
	}

// Sets up a multicast send/receive socket for the specified
// port on the interface specified by the IP address intfAddr.
static int SetupSocket(struct sockaddr *intfAddr, mDNSIPPort port, int interfaceIndex, int *sktPtr)
	{
	int err = 0;
	static const int kOn = 1;
	static const int kIntTwoFiveFive = 255;
	static const unsigned char kByteTwoFiveFive = 255;
	
	(void) interfaceIndex;	// Unused
	assert(intfAddr != NULL);
	assert(sktPtr != NULL);
	assert(*sktPtr == -1);

	// Open the socket...
	if       (intfAddr->sa_family == AF_INET) *sktPtr = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
#ifdef mDNSIPv6Support
	else if (intfAddr->sa_family == AF_INET6) *sktPtr = socket(PF_INET6, SOCK_DGRAM, IPPROTO_UDP);
#endif
	else return EINVAL;

	if (*sktPtr < 0) { err = errno; perror("socket"); }

	// ... with a shared UDP port
	if (err == 0)
		{
		#if defined(SO_REUSEPORT)
#if (OEM_TARGET_OS != VXWORKS)
			err = setsockopt(*sktPtr, SOL_SOCKET, SO_REUSEPORT, &kOn, sizeof(kOn));
#else
			err = setsockopt(*sktPtr, SOL_SOCKET, SO_REUSEPORT, (char *)&kOn, sizeof(kOn));
#endif /* (OEM_TARGET_OS != VXWORKS) */

		#elif defined(SO_REUSEADDR)
#if (OEM_TARGET_OS != VXWORKS)
			err = setsockopt(*sktPtr, SOL_SOCKET, SO_REUSEADDR, &kOn, sizeof(kOn));
#else
			err = setsockopt(*sktPtr, SOL_SOCKET, SO_REUSEADDR, (char *)&kOn, sizeof(kOn));
#endif /* (OEM_TARGET_OS != VXWORKS) */
		#else
			#error This platform has no way to avoid address busy errors on multicast.
		#endif
		if (err < 0) { err = errno; perror("setsockopt - SO_REUSExxxx"); }
		}

	// We want to receive destination addresses and interface identifiers.
	if (intfAddr->sa_family == AF_INET)
		{
		struct ip_mreq imr;
		struct sockaddr_in bindAddr;
		if (err == 0)
			{
			#if defined(IP_PKTINFO)									// Linux
				err = setsockopt(*sktPtr, IPPROTO_IP, IP_PKTINFO, &kOn, sizeof(kOn));
				if (err < 0) { err = errno; perror("setsockopt - IP_PKTINFO"); }
			#elif defined(IP_RECVDSTADDR) || defined(IP_RECVIF)		// BSD and Solaris
				#if defined(IP_RECVDSTADDR)
#if (OEM_TARGET_OS != VXWORKS)
					err = setsockopt(*sktPtr, IPPROTO_IP, IP_RECVDSTADDR, &kOn, sizeof(kOn));
#else
					err = setsockopt(*sktPtr, IPPROTO_IP, IP_RECVDSTADDR, (char *)&kOn, sizeof(kOn));
#endif /* (OEM_TARGET_OS != VXWORKS) */
					if (err < 0) { err = errno; perror("setsockopt - IP_RECVDSTADDR"); }
				#endif
				#if defined(IP_RECVIF)
					if (err == 0)
						{
						err = setsockopt(*sktPtr, IPPROTO_IP, IP_RECVIF, &kOn, sizeof(kOn));
						if (err < 0) { err = errno; perror("setsockopt - IP_RECVIF"); }
						}
				#endif
			#else
				#warning This platform has no way to get the destination interface information -- will only work for single-homed hosts
			#endif
			}

		// Add multicast group membership on this interface
		if (err == 0)
			{
			imr.imr_multiaddr.s_addr = AllDNSLinkGroup.NotAnInteger;
			imr.imr_interface        = ((struct sockaddr_in*)intfAddr)->sin_addr;
#if (OEM_TARGET_OS != VXWORKS)
			err = setsockopt(*sktPtr, IPPROTO_IP, IP_ADD_MEMBERSHIP, &imr, sizeof(imr));
#else
			err = setsockopt(*sktPtr, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&imr, sizeof(imr));
#endif /* (OEM_TARGET_OS != VXWORKS) */
			if (err < 0) { err = errno; perror("setsockopt - IP_ADD_MEMBERSHIP"); }
			}

		// Specify outgoing interface too
		if (err == 0)
			{
#if (OEM_TARGET_OS != VXWORKS)
			err = setsockopt(*sktPtr, IPPROTO_IP, IP_MULTICAST_IF, &((struct sockaddr_in*)intfAddr)->sin_addr, sizeof(struct in_addr));
#else
			err = setsockopt(*sktPtr, IPPROTO_IP, IP_MULTICAST_IF, (char *)&((struct sockaddr_in*)intfAddr)->sin_addr, sizeof(struct in_addr));
#endif /* (OEM_TARGET_OS != VXWORKS) */
			if (err < 0) { err = errno; perror("setsockopt - IP_MULTICAST_IF"); }
			}

		// Per the mDNS spec, send unicast packets with TTL 255
		if (err == 0)
			{
#if (OEM_TARGET_OS != VXWORKS)
			err = setsockopt(*sktPtr, IPPROTO_IP, IP_TTL, &kIntTwoFiveFive, sizeof(kIntTwoFiveFive));
#else
			err = setsockopt(*sktPtr, IPPROTO_IP, IP_TTL, (char *)&kIntTwoFiveFive, sizeof(kIntTwoFiveFive));
#endif /* (OEM_TARGET_OS != VXWORKS) */
			if (err < 0) { err = errno; perror("setsockopt - IP_TTL"); }
			}

		// and multicast packets with TTL 255 too
		// There's some debate as to whether IP_MULTICAST_TTL is an int or a byte so we just try both.
		if (err == 0)
			{
#if (OEM_TARGET_OS != VXWORKS)
			err = setsockopt(*sktPtr, IPPROTO_IP, IP_MULTICAST_TTL, &kByteTwoFiveFive, sizeof(kByteTwoFiveFive));
#else
			err = setsockopt(*sktPtr, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&kByteTwoFiveFive, sizeof(kByteTwoFiveFive));
#endif /* (OEM_TARGET_OS != VXWORKS) */
			if (err < 0 && errno == EINVAL)
#if (OEM_TARGET_OS != VXWORKS)
				err = setsockopt(*sktPtr, IPPROTO_IP, IP_MULTICAST_TTL, &kIntTwoFiveFive, sizeof(kIntTwoFiveFive));
#else
				err = setsockopt(*sktPtr, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&kIntTwoFiveFive, sizeof(kIntTwoFiveFive));
#endif /* (OEM_TARGET_OS != VXWORKS) */
			if (err < 0) { err = errno; perror("setsockopt - IP_MULTICAST_TTL"); }
			}

		// And start listening for packets
		if (err == 0)
			{
			bindAddr.sin_family      = AF_INET;
			bindAddr.sin_port        = port.NotAnInteger;
			bindAddr.sin_addr.s_addr = INADDR_ANY; // Want to receive multicasts AND unicasts on this socket
			err = bind(*sktPtr, (struct sockaddr *) &bindAddr, sizeof(bindAddr));
			if (err < 0) { err = errno; perror("bind"); fflush(stderr); }
			}
#if (OEM_TARGET_OS == VXWORKS)
		{
		    // Turn off loopback so we do not receive multicast packets that we send.
		    unsigned char optionByte = 0;
		    err = setsockopt( *sktPtr, IPPROTO_IP, IP_MULTICAST_LOOP, &optionByte, sizeof( optionByte ) );
		    check_errno( err, errno );
		}
#endif /* OEM_TARGET_OS == VXWORKS */
		} // endif (intfAddr->sa_family == AF_INET)

#ifdef mDNSIPv6Support
	else if (intfAddr->sa_family == AF_INET6)
		{
		struct ipv6_mreq imr6;
		struct sockaddr_in6 bindAddr6;
		if (err == 0)
			{
			#if defined(IPV6_PKTINFO)
				err = setsockopt(*sktPtr, IPPROTO_IPV6, IPV6_PKTINFO, &kOn, sizeof(kOn));
				if (err < 0) { err = errno; perror("setsockopt - IPV6_PKTINFO"); }
			#else
				#warning This platform has no way to get the destination interface information for IPv6 -- will only work for single-homed hosts
			#endif
			}

		// Add multicast group membership on this interface
		if (err == 0)
			{
			imr6.ipv6mr_multiaddr       = *(const struct in6_addr*)&AllDNSLinkGroupv6;
			imr6.ipv6mr_interface       = interfaceIndex;
			err = setsockopt(*sktPtr, IPPROTO_IPV6, IPV6_JOIN_GROUP, &imr6, sizeof(imr6));
			if (err < 0)
				{
				err = errno;
				verbosedebugf("IPV6_JOIN_GROUP %.16a on %d failed.\n", &imr6.ipv6mr_multiaddr, imr6.ipv6mr_interface);
				perror("setsockopt - IPV6_JOIN_GROUP");
				}
			}

		// Specify outgoing interface too
		if (err == 0)
			{
			u_int	multicast_if = interfaceIndex;
			err = setsockopt(*sktPtr, IPPROTO_IPV6, IPV6_MULTICAST_IF, &multicast_if, sizeof(multicast_if));
			if (err < 0) { err = errno; perror("setsockopt - IPV6_MULTICAST_IF"); }
			}

		// We want to receive only IPv6 packets on this socket.
		// Without this option, we may get IPv4 addresses as mapped addresses.
		if (err == 0)
			{
			err = setsockopt(*sktPtr, IPPROTO_IPV6, IPV6_V6ONLY, &kOn, sizeof(kOn));
			if (err < 0) { err = errno; perror("setsockopt - IPV6_V6ONLY"); }
			}

		// Per the mDNS spec, send unicast packets with TTL 255
		if (err == 0)
			{
			err = setsockopt(*sktPtr, IPPROTO_IPV6, IPV6_UNICAST_HOPS, &kIntTwoFiveFive, sizeof(kIntTwoFiveFive));
			if (err < 0) { err = errno; perror("setsockopt - IPV6_UNICAST_HOPS"); }
			}

		// and multicast packets with TTL 255 too
		// There's some debate as to whether IPV6_MULTICAST_HOPS is an int or a byte so we just try both.
		if (err == 0)
			{
			err = setsockopt(*sktPtr, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, &kByteTwoFiveFive, sizeof(kByteTwoFiveFive));
			if (err < 0 && errno == EINVAL)
				err = setsockopt(*sktPtr, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, &kIntTwoFiveFive, sizeof(kIntTwoFiveFive));
			if (err < 0) { err = errno; perror("setsockopt - IPV6_MULTICAST_HOPS"); }
			}

		// And start listening for packets
		if (err == 0)
			{
			mDNSPlatformMemZero(&bindAddr6, sizeof(bindAddr6));
			bindAddr6.sin6_len         = sizeof(bindAddr6);
			bindAddr6.sin6_family      = AF_INET6;
			bindAddr6.sin6_port        = port.NotAnInteger;
			bindAddr6.sin6_flowinfo    = 0;
//			bindAddr6.sin6_addr.s_addr = IN6ADDR_ANY_INIT; // Want to receive multicasts AND unicasts on this socket
			bindAddr6.sin6_scope_id    = 0;
			err = bind(*sktPtr, (struct sockaddr *) &bindAddr6, sizeof(bindAddr6));
			if (err < 0) { err = errno; perror("bind"); fflush(stderr); }
			}
		} // endif (intfAddr->sa_family == AF_INET6)
#endif

	// Set the socket to non-blocking.
	if (err == 0)
		{
#if (OEM_TARGET_OS == VXWORKS)
		int one = 1;
		err = ioctl(*sktPtr, FIONBIO, (int )&one);
		if (err < 0)
			{
			err = errno;
			}
#else
		err = fcntl(*sktPtr, F_GETFL, 0);
		if (err < 0) err = errno;
		else
			{
			err = fcntl(*sktPtr, F_SETFL, err | O_NONBLOCK);
			if (err < 0) err = errno;
			}
#endif /* (OEM_TARGET_OS == VXWORKS) */
		}

	// Clean up
	if (err != 0 && *sktPtr != -1) { assert(close(*sktPtr) == 0); *sktPtr = -1; }
	assert( (err == 0) == (*sktPtr != -1) );
	return err;
	}

// Creates a PosixNetworkInterface for the interface whose IP address is
// intfAddr and whose name is intfName and registers it with mDNS core.
static int SetupOneInterface(mDNS *const m, struct sockaddr *intfAddr, const char *intfName)
	{
	int err = 0;
	PosixNetworkInterface *intf;
	PosixNetworkInterface *alias = NULL;

	assert(m != NULL);
	assert(intfAddr != NULL);
	assert(intfName != NULL);

	// Allocate the interface structure itself.
	intf = malloc(sizeof(*intf));
	if (intf == NULL) { assert(0); err = ENOMEM; }

	// And make a copy of the intfName.
	if (err == 0)
		{
		intf->intfName = strdup(intfName);
		if (intf->intfName == NULL) { assert(0); err = ENOMEM; }
		}

	if (err == 0)
		{
		// Set up the fields required by the mDNS core.
		SockAddrTomDNSAddr(intfAddr, &intf->coreIntf.ip, NULL);
		intf->coreIntf.Advertise = m->AdvertiseLocalAddresses;
		intf->coreIntf.TxAndRx   = mDNStrue;

		// Set up the extra fields in PosixNetworkInterface.
		assert(intf->intfName != NULL);         // intf->intfName already set up above
#if (OEM_TARGET_OS != VXWORKS)
		intf->index                = if_nametoindex(intf->intfName);
#else
		/* XeroxOPB: only have one interface so could return 0 */
		intf->index                = opb_if_nametoindex((char *)intf->intfName);
#endif /* (OEM_TARGET_OS != VXWORKS) */
		intf->multicastSocket      = -1;
		intf->multicastSocketv6    = -1;
		alias                      = SearchForInterfaceByName(m, intf->intfName);
		if (alias == NULL) alias   = intf;
		intf->coreIntf.InterfaceID = (mDNSInterfaceID)alias;

		if (alias != intf)
			debugf("SetupOneInterface: %s %#a is an alias of %#a", intfName, &intf->coreIntf.ip, &alias->coreIntf.ip);
		}

	// Set up the multicast socket
	if (err == 0)
		{
		if (alias->multicastSocket == -1 && intfAddr->sa_family == AF_INET)
			err = SetupSocket(intfAddr, MulticastDNSPort, intf->index, &alias->multicastSocket);
#ifdef mDNSIPv6Support
		else if (alias->multicastSocketv6 == -1 && intfAddr->sa_family == AF_INET6)
			err = SetupSocket(intfAddr, MulticastDNSPort, intf->index, &alias->multicastSocketv6);
#endif
		}

	// The interface is all ready to go, let's register it with the mDNS core.
	if (err == 0)
		err = mDNS_RegisterInterface(m, &intf->coreIntf);

	// Clean up.
	if (err == 0)
		{
		num_registered_interfaces++;
		debugf("SetupOneInterface: %s %#a Registered", intf->intfName, &intf->coreIntf.ip);
		if (gMDNSPlatformPosixVerboseLevel > 0)
			fprintf(stderr, "Registered interface %s\n", intf->intfName);
		}
	else
		{
		// Use intfName instead of intf->intfName in the next line to avoid dereferencing NULL.
		debugf("SetupOneInterface: %s %#a failed to register %d", intfName, &intf->coreIntf.ip, err);
		if (intf) { FreePosixNetworkInterface(intf); intf = NULL; }
		}

	assert( (err == 0) == (intf != NULL) );

	return err;
	}

static int SetupInterfaceList(mDNS *const m)
	{
	mDNSBool        foundav4       = mDNSfalse;
	int             err            = 0;
#if (OEM_TARGET_OS != VXWORKS)
	struct ifi_info *intfList      = get_ifi_info(AF_INET, mDNStrue);
#else
	struct ifi_info *intfList      = NULL;
	struct ifi_info *thisIntf      = NULL;
#endif /* (OEM_TARGET_OS != VXWORKS) */
	struct ifi_info *firstLoopback = NULL;

	assert(m != NULL);
	debugf("SetupInterfaceList");

#if (OEM_TARGET_OS == VXWORKS)
	// OPB: need to allow time for the interface to synch with the config
	// saved in NCL. get the interface info, and make sure that the 
	//current IP address is in.
	while (1) {
	    boolean matchFound = FALSE;
	    intfList = get_ifi_info(AF_INET, mDNStrue);
	    if (intfList == NULL) {
		err = ENOENT;
		break;
	    }
	    if (m->p->trackedNCLvars.ipAddr.s_addr == 0) {
		break;  // accept the list as-is
	    }

	    // make sure our IP address is in the list
	    thisIntf = intfList;
	    while (thisIntf != NULL) {
		if ((thisIntf->ifi_addr->sa_family == AF_INET) &&
		    (thisIntf->ifi_flags & IFF_UP) &&
		    (((struct sockaddr_in *)(thisIntf->ifi_addr))->sin_addr.s_addr == m->p->trackedNCLvars.ipAddr.s_addr)) {
		    matchFound = TRUE;   // it's in the list
		    break;
		}
		thisIntf = thisIntf->ifi_next;  // keep looking
	    }
	    if (matchFound) break;  // done
	    // not yet, cleanup the list and pause a moment to allow the 
	    // interface configuration process to catch up.. seems to take ~15-20+
	    // seconds, but keep checking sooner
	    free_ifi_info(intfList);
	    taskDelay(5 * sysClkRateGet());  // wait 5 secs.
	}
#endif /* OEM_TARGET_OS == VXWORKS */
	if (intfList == NULL) err = ENOENT;

#ifdef mDNSIPv6Support
	if (err == 0)		/* Link the IPv6 list to the end of the IPv4 list */
		{
		struct ifi_info **p = &intfList;
		while (*p) p = &(*p)->ifi_next;
		*p = get_ifi_info(AF_INET6, mDNStrue);
		}
#endif

	if (err == 0)
		{
		struct ifi_info *i = intfList;
		while (i)
			{
			if (     ((i->ifi_addr->sa_family == AF_INET)
#ifdef mDNSIPv6Support
					  || (i->ifi_addr->sa_family == AF_INET6)
#endif
				) &&  (i->ifi_flags & IFF_UP) && !(i->ifi_flags & IFF_POINTOPOINT) )
				{
				if (i->ifi_flags & IFF_LOOPBACK)
					{
					if (firstLoopback == NULL)
						firstLoopback = i;
					}
				else
					{
#if (OEM_TARGET_OS == VXWORKS)
		    // ensure that only the configuration matching
		    // our NCL view of the "current" IP address is used
					if (m->p->trackedNCLvars.ipAddr.s_addr && (((struct sockaddr_in *)(thisIntf->ifi_addr))->sin_addr.s_addr == m->p->trackedNCLvars.ipAddr.s_addr)) {
#endif /* OEM_TARGET_OS == VXWORKS */
					if (SetupOneInterface(m, i->ifi_addr, i->ifi_name) == 0)
						if (i->ifi_addr->sa_family == AF_INET)
							foundav4 = mDNStrue;
#if (OEM_TARGET_OS == VXWORKS)
					}
#endif
					}
				}
			i = i->ifi_next;
			}

		// If we found no normal interfaces but we did find a loopback interface, register the
		// loopback interface.  This allows self-discovery if no interfaces are configured.
		// Temporary workaround: Multicast loopback on IPv6 interfaces appears not to work.
		// In the interim, we skip loopback interface only if we found at least one v4 interface to use
		// if ( (m->HostInterfaces == NULL) && (firstLoopback != NULL) )
		if ( !foundav4 && firstLoopback )
			(void) SetupOneInterface(m, firstLoopback->ifi_addr, firstLoopback->ifi_name);
		}

	// Clean up.
	if (intfList != NULL) free_ifi_info(intfList);
	return err;
	}

#if (OEM_TARGET_OS == VXWORKS)
//===========================================================================================================================
//	ProcessCommandUpdateIP
//===========================================================================================================================

static void	ProcessCommandUpdateIP( mDNS *inMDNS )
{
	//	debug_printf( kACPDebugLevelVerboseInfo, DEBUG_NAME "processing pipe update IP command\n" );
	
	// Tear down the existing interfaces and set up new ones using the new IP info.
	
	//	mDNSPlatformLock( inMDNS );
	
	// the work is done in the Responder's main select loop
	// since service (de)registration is defined in the Responder.c module
	
	//	mDNSPlatformUnlock( inMDNS );
	
	// Bump the config ID so the main processing loop detects the configuration change.
	
	++inMDNS->p->configID;
}

//===========================================================================================================================
//	TearDownCommandPipe
//===========================================================================================================================

static mStatus	TearDownCommandPipe( mDNS * const inMDNS )
{
    mStatus		err;
	
	require_action_quiet( inMDNS->p->commandPipe != ERROR, exit, err = mStatus_NotInitializedErr );
	
    if (inMDNS->p && inMDNS->p->commandPipe != ERROR) {
	close( inMDNS->p->commandPipe );
	inMDNS->p->commandPipe = ERROR;
    }
	
    err = mStatus_NoError;
	
exit:
    return( err );
}

//===========================================================================================================================
//	SetupCommandPipe
//===========================================================================================================================

static mStatus	SetupCommandPipe( mDNS * const inMDNS )
{
	mStatus		err;
	
	// Clean up any leftover command pipe.
	
	TearDownCommandPipe( inMDNS );
	
	// Create the pipe device and open it.
	
	pipeDevCreate( kMDNSPipeName, kMDNSPipeMessageQueueSize, kMDNSPipeMessageSize );
	
	inMDNS->p->commandPipe = open( kMDNSPipeName, O_RDWR, 0 );
	require_errno_action( inMDNS->p->commandPipe, errno, exit, err = mStatus_NotInitializedErr );
	
	err = mStatus_NoError;
	
exit:
	return( err );
}

//===========================================================================================================================
//	SendCommand
//===========================================================================================================================

static mStatus	SendCommand( mDNS * const inMDNS, MDNSPipeCommandCode inCommandCode )
{
	mStatus		err;
	
#ifdef MDNS_DEBUG
	lprintf("SendCommand: %d\n",inCommandCode);
#endif /* MDNS_DEBUG */
	require_action_quiet( inMDNS->p->commandPipe != ERROR, exit, err = mStatus_NotInitializedErr );
	
	err = write( inMDNS->p->commandPipe, &inCommandCode, sizeof( inCommandCode ) );
	//	require_errno( err, errno, exit );
	
	err = mStatus_NoError;

exit:
	return( err );
}

//===========================================================================================================================
//	ProcessCommand
//===========================================================================================================================

static mStatus	ProcessCommand( mDNS * const inMDNS )
{
	mStatus					err;
	MDNSPipeCommandCode		commandCode;
	
#ifdef MDNS_DEBUG
	lprintf("pipe ProcessCommand\n");
#endif /* MDNS_DEBUG */
	//	require_action( inMDNS->p->commandPipe != ERROR, exit, err = kACPNotInitializedErr );
	
	// Read the command code from the pipe and dispatch it.
	
	err = read( inMDNS->p->commandPipe, &commandCode, sizeof( commandCode ) );
	//	require_errno( err, errno, exit );
		
	switch( commandCode )
	{
		case kMDNSPipeCommandCodeUpdateIP:
			ProcessCommandUpdateIP( inMDNS );
			break;
		
		case kMDNSPipeCommandCodeIFLinkChange:
			// accepting this command results in returning 
			// no error status, which then will cause a
			// deregister/ reregister in the mDNSResponder 
			// main loop
			break;
		
		//the following are currently unsupported
		case kMDNSPipeCommandCodeTimer:
		case kMDNSPipeCommandCodeQuit:
		case kMDNSPipeCommandCodeNameCollision:
		default:
			//			debug_printf( kACPDebugLevelError, DEBUG_NAME "unknown pipe command code (code=0x%08X)\n", commandCode );
			//			err = kACPCommandNotHandledErr;
			err = mStatus_UnsupportedErr;
			goto exit;
			break;
	}
	err = mStatus_NoError;

exit:
	return( err );
}
#endif /* (OEM_TARGET_OS == VXWORKS) */

// mDNS core calls this routine to initialise the platform-specific data.
mDNSexport mStatus mDNSPlatformInit(mDNS *const m)
	{
#if (OEM_TARGET_OS != VXWORKS)
	int err;
#else
    int err = 0;
#endif /* (OEM_TARGET_OS != VXWORKS) */
	assert(m != NULL);

	// Tell mDNS core the names of this machine.

	// Set up the nice label
	m->nicelabel.c[0] = 0;
#if (OEM_TARGET_OS != VXWORKS)
	GetUserSpecifiedFriendlyComputerName(&m->nicelabel);
	if (m->nicelabel.c[0] == 0) MakeDomainLabelFromLiteralString(&m->nicelabel, "Macintosh");
#else
	GetUserSpecifiedFriendlyComputerName(m, &m->nicelabel);
#endif /* (OEM_TARGET_OS != VXWORKS) */

	// Set up the RFC 1034-compliant label
	m->hostlabel.c[0] = 0;
#if (OEM_TARGET_OS != VXWORKS)
	GetUserSpecifiedRFC1034ComputerName(&m->hostlabel);
	if (m->hostlabel.c[0] == 0) MakeDomainLabelFromLiteralString(&m->hostlabel, "Macintosh");
#else
	GetUserSpecifiedRFC1034ComputerName(m, &m->hostlabel);
	if ((m->nicelabel.c[0] == 0) || (m->hostlabel.c[0] == 0)) {
	    MakeDomainLabelFromLiteralString(&m->hostlabel, m->p->pRichTextHostName);
	    ConvertUTF8PstringToRFC1034HostLabel(m->nicelabel.c, &m->hostlabel);
	}
#endif /* (OEM_TARGET_OS != VXWORKS) */

	mDNS_GenerateFQDN(m);

#if (OEM_TARGET_OS == VXWORKS)
	if (err == 0) {
	    err = SetupCommandPipe( m );
	}
#endif /* (OEM_TARGET_OS == VXWORKS) */
	// Tell mDNS core about the network interfaces on this machine.
#if  (OEM_TARGET_OS == VXWORKS)
	if (err == 0) {
#endif /* (OEM_TARGET_OS == VXWORKS) */
	err = SetupInterfaceList(m);
#if  (OEM_TARGET_OS == VXWORKS)
	}
#endif /* (OEM_TARGET_OS == VXWORKS) */

	// We don't do asynchronous initialization on the Posix platform, so by the time
	// we get here the setup will already have succeeded or failed.  If it succeeded,
	// we should just call mDNSCoreInitComplete() immediately.
	if (err == 0)
		mDNSCoreInitComplete(m, mStatus_NoError);

	return PosixErrorToStatus(err);
	}

// mDNS core calls this routine to clean up the platform-specific data.
// In our case all we need to do is to tear down every network interface.
mDNSexport void mDNSPlatformClose(mDNS *const m)
	{
	assert(m != NULL);
	ClearInterfaceList(m);
#if (OEM_TARGET_OS == VXWORKS)
	TearDownCommandPipe(m);
#endif /* (OEM_TARGET_OS == VXWORKS) */
	}

extern mStatus mDNSPlatformPosixRefreshInterfaceList(mDNS *const m)
	{
	int err;
	ClearInterfaceList(m);
	err = SetupInterfaceList(m);
	return PosixErrorToStatus(err);
	}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Locking
#endif

// On the Posix platform, locking is a no-op because we only ever enter
// mDNS core on the main thread.

// mDNS core calls this routine when it wants to prevent
// the platform from reentering mDNS core code.
mDNSexport void    mDNSPlatformLock   (const mDNS *const m)
	{
#if (OEM_TARGET_OS != VXWORKS)
	(void) m;	// Unused
#else
	check( m->p->lockID );
	
	if( m->p->lockID )
	    {
	    semTake( m->p->lockID, WAIT_FOREVER );
	    }
#endif /* (OEM_TARGET_OS == VXWORKS) */
	}

// mDNS core calls this routine when it release the lock taken by
// mDNSPlatformLock and allow the platform to reenter mDNS core code.
mDNSexport void    mDNSPlatformUnlock (const mDNS *const m)
	{
#if (OEM_TARGET_OS != VXWORKS)
	(void) m;	// Unused
#else
	check( m->p->lockID );
	
	if( m->p->lockID )
	    {
	    semGive( m->p->lockID);
	    }
#endif /* (OEM_TARGET_OS == VXWORKS) */
	}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Strings
#endif

// mDNS core calls this routine to copy C strings.
// On the Posix platform this maps directly to the ANSI C strcpy.
mDNSexport void    mDNSPlatformStrCopy(const void *src,       void *dst)
	{
	strcpy((char *)dst, (char *)src);
	}

// mDNS core calls this routine to get the length of a C string.
// On the Posix platform this maps directly to the ANSI C strlen.
mDNSexport mDNSu32  mDNSPlatformStrLen (const void *src)
	{
	return strlen((char*)src);
	}

// mDNS core calls this routine to copy memory.
// On the Posix platform this maps directly to the ANSI C memcpy.
mDNSexport void    mDNSPlatformMemCopy(const void *src,       void *dst, mDNSu32 len)
	{
	memcpy(dst, src, len);
	}

// mDNS core calls this routine to test whether blocks of memory are byte-for-byte
// identical. On the Posix platform this is a simple wrapper around ANSI C memcmp.
mDNSexport mDNSBool mDNSPlatformMemSame(const void *src, const void *dst, mDNSu32 len)
	{
	return memcmp(dst, src, len) == 0;
	}

// mDNS core calls this routine to clear blocks of memory.
// On the Posix platform this is a simple wrapper around ANSI C memset.
mDNSexport void    mDNSPlatformMemZero(                       void *dst, mDNSu32 len)
	{
	memset(dst, 0, len);
	}

mDNSexport void *  mDNSPlatformMemAllocate(mDNSu32 len) { return(malloc(len)); }
mDNSexport void    mDNSPlatformMemFree    (void *mem)   { free(mem); }

mDNSexport mDNSs32  mDNSPlatformOneSecond = 1024;

mDNSexport mStatus mDNSPlatformTimeInit(mDNSs32 *timenow)
	{
	// No special setup is required on Posix -- we just use gettimeofday();
	// This is not really safe, because gettimeofday can go backwards if the user manually changes the date or time
	// We should find a better way to do this
	*timenow = mDNSPlatformTimeNow();
	return(mStatus_NoError);
	}

mDNSexport mDNSs32  mDNSPlatformTimeNow()
	{
	struct timeval tv;
#if (OEM_TARGET_OS != VXWORKS)
	gettimeofday(&tv, NULL);
#else
        {
	struct timespec tp;
	clock_gettime(CLOCK_REALTIME, &tp);
	tv.tv_sec = tp.tv_sec;
	tv.tv_usec = tp.tv_nsec / 1000;
        }
#endif /* (OEM_TARGET_OS != VXWORKS) */
	// tv.tv_sec is seconds since 1st January 1970 (GMT, with no adjustment for daylight savings time)
	// tv.tv_usec is microseconds since the start of this second (i.e. values 0 to 999999)
	// We use the lower 22 bits of tv.tv_sec for the top 22 bits of our result
	// and we multiply tv.tv_usec by 16 / 15625 to get a value in the range 0-1023 to go in the bottom 10 bits.
	// This gives us a proper modular (cyclic) counter that has a resolution of roughly 1ms (actually 1/1024 second)
	// and correctly cycles every 2^22 seconds (4194304 seconds = approx 48 days).
	return( (tv.tv_sec << 10) | (tv.tv_usec * 16 / 15625) );
	}

mDNSexport void mDNSPosixGetFDSet(mDNS *const m, int *nfds, fd_set *readfds, struct timeval *timeout)
	{
	mDNSs32 ticks;
	struct timeval interval;

	// 1. Call mDNS_Execute() to let mDNSCore do what it needs to do
	mDNSs32 nextevent = mDNS_Execute(m);

	// 2. Build our list of active file descriptors
	PosixNetworkInterface *info = (PosixNetworkInterface *)(m->HostInterfaces);
	while (info)
		{
		if (info->multicastSocket != -1)
			{
			if (*nfds < info->multicastSocket + 1)
				*nfds = info->multicastSocket + 1;
			FD_SET(info->multicastSocket, readfds);
			}
		if (info->multicastSocketv6 != -1)
			{
			if (*nfds < info->multicastSocketv6 + 1)
				*nfds = info->multicastSocketv6 + 1;
			FD_SET(info->multicastSocketv6, readfds);
			}
		info = (PosixNetworkInterface *)(info->coreIntf.next);
		}
#if (OEM_TARGET_OS == VXWORKS)
	// Add the command pipe to the read set.

	if (m && m->p && m->p->commandPipe != ERROR)
		{
		FD_SET( m->p->commandPipe, readfds );
		if(  *nfds < m->p->commandPipe + 1)
			{
			*nfds = m->p->commandPipe + 1;
			}
		}
#endif
	// 3. Calculate the time remaining to the next scheduled event (in struct timeval format)
	ticks = nextevent - mDNSPlatformTimeNow();
	if (ticks < 1) ticks = 1;
	interval.tv_sec  = ticks >> 10;						// The high 22 bits are seconds
	interval.tv_usec = ((ticks & 0x3FF) * 15625) / 16;	// The low 10 bits are 1024ths

	// 4. If client's proposed timeout is more than what we want, then reduce it
#if (OEM_TARGET_OS == VXWORKS)
	// allow poll mode timer for select if 0, but otherwise force timer
	// to just over a system tick to prevent poll mode
#define JUSTOVERONETICK (1000000/sysClkRateGet() + 1)
	if (interval.tv_sec == 0)
	{
	    if (interval.tv_usec && (interval.tv_usec < JUSTOVERONETICK))
		interval.tv_usec = JUSTOVERONETICK;
	}
#endif /* (OEM_TARGET_OS == VXWORKS) */
	if (timeout->tv_sec > interval.tv_sec ||
		(timeout->tv_sec == interval.tv_sec && timeout->tv_usec > interval.tv_usec))
		*timeout = interval;
	}

#if (OEM_TARGET_OS != VXWORKS)
mDNSexport void mDNSPosixProcessFDSet(mDNS *const m, fd_set *readfds)
#else
mDNSexport int mDNSPosixProcessFDSet(mDNS *const m, int selectresult, fd_set *readfds)
#endif /* (OEM_TARGET_OS != VXWORKS) */
	{
	PosixNetworkInterface *info;
	assert(m       != NULL);
	assert(readfds != NULL);
#if (OEM_TARGET_OS == VXWORKS)
	// optimization: don't bother checking the readfs mask if select
	// returned 0 (select timeout return)
	if (selectresult > 0)
		{
#endif /* (OEM_TARGET_OS == VXWORKS) */
	info = (PosixNetworkInterface *)(m->HostInterfaces);
	while (info)
		{
		if (info->multicastSocket != -1 && FD_ISSET(info->multicastSocket, readfds))
			{
			FD_CLR(info->multicastSocket, readfds);
			SocketDataReady(m, info, info->multicastSocket);
			}
		if (info->multicastSocketv6 != -1 && FD_ISSET(info->multicastSocketv6, readfds))
			{
			FD_CLR(info->multicastSocketv6, readfds);
			SocketDataReady(m, info, info->multicastSocketv6);
			}
		info = (PosixNetworkInterface *)(info->coreIntf.next);
		}
#if (OEM_TARGET_OS == VXWORKS)
	// Check for a pending command and process it.

	if(m && m->p && (m->p->commandPipe != ERROR) && FD_ISSET( m->p->commandPipe, readfds)) {
	    mStatus status;
	    if ((status = ProcessCommand( m )) == mStatus_NoError)
		return (-1);
	}
#if (OEM_TARGET_OS == VXWORKS)
		}
#endif /* (OEM_TARGET_OS == VXWORKS) */
	
	return (0);
#endif /* (OEM_TARGET_OS == VXWORKS) */
	}

#if (OEM_TARGET_OS == VXWORKS)
boolean
OPBSetPSDeviceOnEnabled(char *deviceParamSet)
{
    bool deviceEnable = FALSE;
    bool deviceOn = FALSE;
    char paramName[XPM_MAX_GROUPNAME_LEN + XPM_MAX_PARAMNAME_LEN + 2];
    char *ptr;

    // get PS parameter set
    strcpy(paramName, deviceParamSet);
    ptr = paramName + strlen(paramName);
    *ptr = '.';
    ptr++;
    strcpy(ptr, "Enabled");
    xpmGetParamBoolVal (XpmHandle_None, paramName, &deviceEnable, ParamAccLvl_System);
    strcpy(ptr, "On");
    xpmGetParamBoolVal (XpmHandle_None, paramName, &deviceOn, ParamAccLvl_System);
    // both must be TRUE for device  to be working
    return (deviceEnable && deviceOn);
}

/*
 * opbAddPascalString
 *
 * add up to 3 substrings to the target string, prepending the length of
 * the string in the 1st character
 */
static void
opbAddPascalString(char *target, char *part1, char *part2, char *part3)
{
    char built[512];
    int builtLen;

    /* prime */
    built[0] = ' ';
    built[1] = '\0';
    /* add parts */
    if (part1 && *part1) {
        strcat(built, part1);
    }
    if (part2 && *part2) {
        strcat(built, part2);
    }
    if (part3 && *part3) {
        strcat(built, part3);
    }
    // ensure the length won't overflow before adding
    if((builtLen = strlen(built + 1)) <= 255) {
	// prefix the length of the string
	built[0] = (char) builtLen;
	// append the built string to the target
	strcat(target, built);
    }
}

/*
 * opbKillNumericSuffix()
 *
 * remove the "-<#>" suffix which was added as an AppleTalk name uniqueness
 * suffix.
 *
 * returns TRUE if a suffix was removed, FALSE otherwise
 */
static int
opbKillNumericSuffix(char *name, int startingChar)
{
    char *ptr, *limit;
    int nosDigits = 0;
 
    if (name == NULL) return (FALSE);

    limit = name + startingChar;
    /* look for the form "\0" */
    if (*limit == '\0') return (FALSE);
    ptr = name + strlen(name) - 1;
    /* look for the form "-<#>" */
    while(ptr > limit) {
	if (isdigit(*ptr)) {
	    ptr--;
	    nosDigits++;
	}
	else {
	    break;
	}
    }
    if ((*ptr != '-') || (nosDigits == 0))
	return (FALSE);
    *ptr = '\0';
    return (TRUE);
}

/*
 * opbEscapeSpecialChars()
 *
 * change the name in place by escaping any "." or "\", but do not overflow
 *
 */
static void
opbEscapeSpecialChars(char *name, int maxLen)
{
    char *end = name + maxLen;
    while (*name) {
	if ((*name == '.') || (*name == '\\')) {
	    int lenMove = strlen(name);
	    if ((name + 1 + lenMove) >= end)
		lenMove--;
	    if (lenMove) memmove((name + 1), name, lenMove);
	    *(name + 1 + lenMove) = '\0';
	    *name = '\\';
	    name++;
	}
	name++;
	if (name >= end) break;
    }
    return;
}

void
examineNvvars(int ignorednclSubject, mDNS_PlatformSupport *platform0, mDNS *m0)
{
    bool IPConfiged;
    struct in_addr ipAddr;
    char hostName[STR_HOST_LENGTH_XPM + 1];
    char deviceName[ID_NAME_LENGTH + 1];
    nuint8 changed = FALSE;
    bool mdnsEnabled;
    static mDNS *m = NULL;
    static  mDNS_PlatformSupport *platform = NULL;
    char printerId[ID_ADDR_LENGTH + 1];
    int port;
    int ipAddrSource = IP_SOURCE_NONE;
    char scratch[STR_HOST_LENGTH + STR_DOMAIN_LENGTH + 1];
    char printerLocation[ID_STR_LENGTH_XPM + 1];
    char modelName[ID_STR_LENGTH + 1];
    int modelNameLen;
    bool needUniquenessSuffix = FALSE;
    bool forceNameSuffix;

#ifdef MDNS_DEBUG
    lprintf("################mDNS examineNvvars##################\n");
#endif /* MDNS_DEBUG */


    if (m == NULL) {
	if (m0 == NULL) {
	    return;  /* not yet called by the main line */
	}
	m = m0;
	platform = platform0;
    }

    // if the printer is shutting down, don't do anything
    if (NEST_stopping) return;

    // unconditional params
    if ((xpmGetParamBoolVal(XpmHandle_None, "IP.IPAddressConfigured", &IPConfiged, ParamAccLvl_System)) != XpmStatus_OK) {
	IPConfiged = FALSE;
    }
    if (platform->trackedNCLvars.IPConfiged != IPConfiged) {
	changed = TRUE;
	platform->trackedNCLvars.IPConfiged = IPConfiged;
    }

    if ((xpmGetParamBoolVal(XpmHandle_None, "Rendezvous.Enabled", &mdnsEnabled, ParamAccLvl_System)) != XpmStatus_OK) {
	mdnsEnabled = FALSE;
    }
    if (platform->trackedNCLvars.mdnsEnabled != mdnsEnabled) {
	changed = TRUE;
	platform->trackedNCLvars.mdnsEnabled = mdnsEnabled;
#if EVENTLOG_OEM_COMPONENT_INCLUDED
	if (mdnsEnabled)
	    eventlog(EVENTLOG_MDNS, EVENTLOG_RUNTIME, "mDNS (Rendezvous) is enabled");
	else
	    eventlog(EVENTLOG_MDNS, EVENTLOG_RUNTIME, "mDNS (Rendezvous) is disabled");
#endif /* EVENTLOG_OEM_COMPONENT_INCLUDED */

    }
    if ('\0' == platform->trackedNCLvars.uniquenessSuffix[0]) {
	if ((xpmGetParamStrVal(XpmHandle_None, "Ethernet.Address", printerId, ID_ADDR_LENGTH, ParamAccLvl_System)) == XpmStatus_OK) {
	    printerId[2]=':';
	    printerId[5]=':';
	    printerId[8]=':';
	    printerId[11]=':';
	    printerId[14]=':';
	    strcpy(platform->trackedNCLvars.uniquenessSuffix, printerId);
	    platform->trackedNCLvars.uniquenessSuffix[17] = '\0';
	}
	else {
	    platform->trackedNCLvars.uniquenessSuffix[0] = '\0'; 
	    //ALERT: consider crash and burn since the uniqueness suffix
	    // is needed.
#if EVENTLOG_OEM_COMPONENT_INCLUDED
	    eventlog(EVENTLOG_MDNS, EVENTLOG_RUNTIME, "Unable to get OEMPrinterID");
#endif /* EVENTLOG_OEM_COMPONENT_INCLUDED */
	}
    }
    if (((xpmGetParamStrVal(XpmHandle_None, "internal.Product", modelName, ID_NAME_LENGTH, ParamAccLvl_System)) != XpmStatus_OK) || (modelName[0] == '\0')){
	strcpy(modelName, "Phaser");  //fill in "something"
    }
    modelNameLen = strlen(modelName);

    // conditional params
    if (IPConfiged) {
	if (((xpmGetParamIntVal(XpmHandle_None, "IP.IPSource", &ipAddrSource, ParamAccLvl_System)) != XpmStatus_OK) || (ipAddrSource == IP_SOURCE_WAITING)) {
		ipAddrSource = IP_SOURCE_NONE;
	}
	if (ipAddrSource == IP_SOURCE_NONE)
		return;
	else {
	    int status = ERROR;
	    if (ipAddrSource != IP_SOURCE_AUTOIP) {
		status = xpmGetParamByteArrayVal(XpmHandle_None, "IP.IPAddress", (char *) &ipAddr.s_addr, sizeof(ipAddr.s_addr), ParamAccLvl_System);
	    }
	    else {
		status = xpmGetParamByteArrayVal(XpmHandle_None, "IP.Self_AssignedIPAddress", (char *) &ipAddr.s_addr, sizeof(ipAddr.s_addr), ParamAccLvl_System);
	    }
	    if (status== XpmStatus_OK) {
		if (platform->trackedNCLvars.ipAddr.s_addr != ipAddr.s_addr) {
		    changed = TRUE;
		    platform->trackedNCLvars.ipAddr.s_addr = ipAddr.s_addr;
		}
	    }
	    else
		return;
	}
	if (((xpmGetParamStrVal (XpmHandle_None, "IP.HostName", hostName, STR_HOST_LENGTH_XPM + 1, ParamAccLvl_System)) == XpmStatus_OK) && (hostName[0] != '\0')) {
	    if ((strcmp(platform->trackedNCLvars.hostName, hostName) != 0) || (strlen(platform->trackedNCLvars.hostName) != strlen(hostName))){
		changed = TRUE;
		strcpy(platform->trackedNCLvars.hostName, hostName);
	    }
	}
	if ((xpmGetParamStrVal(XpmHandle_None, "internal.PrinterUniqueName", deviceName, ID_NAME_LENGTH,ParamAccLvl_System) != XpmStatus_OK) || (deviceName[0] == '\0')) {
	    strcpy(deviceName, modelName);
	}

	// pick a service root name
	if ((xpmGetParamBoolVal(XpmHandle_None, "Rendezvous.ForceNameUniqueness", &forceNameSuffix, ParamAccLvl_System)) != XpmStatus_OK) {
	    forceNameSuffix = FALSE;
	}
	if (forceNameSuffix == FALSE) {
	    int deviceNameLen;
	    int before = strlen(deviceName);
	    opbKillNumericSuffix(deviceName, 0);
	    // if the name had a uniqueness suffix added OR if the 
	    // name was still == modelName
	    if (forceNameSuffix || 
		((deviceNameLen = strlen(deviceName)) != before) || 
		(((strncmp(deviceName, modelName, modelNameLen)) == 0) && (deviceNameLen == modelNameLen)))
		needUniquenessSuffix = TRUE;
	}

	// OPBU62401 - escape the forbidden "." and "\"
	opbEscapeSpecialChars(deviceName, sizeof(deviceName));
	if ((strcmp(platform->trackedNCLvars.deviceName, deviceName) != 0) || (strlen(platform->trackedNCLvars.deviceName) != strlen(deviceName)) || (platform->trackedNCLvars.needUniquenessSuffix != needUniquenessSuffix)) {
	    changed = TRUE;
	    strcpy(platform->trackedNCLvars.deviceName, deviceName);
	    platform->trackedNCLvars.needUniquenessSuffix = needUniquenessSuffix;
	}
    }

    //  <service suffix>-<nest device> (last 3 chars Enet H/W addr)
    platform->trackedNCLvars.perDeviceServiceRoot[0] ='\0';
    if (platform->trackedNCLvars.deviceName)
	strcpy(platform->trackedNCLvars.perDeviceServiceRoot, platform->trackedNCLvars.deviceName);

    //ALERT: Apple Rendezvous kitchen REALLY lobbied for removing this
    // the future PrintCenter UI "should" fix this
    if (needUniquenessSuffix && platform->trackedNCLvars.uniquenessSuffix) {
	strcat(platform->trackedNCLvars.perDeviceServiceRoot, " (");
	strcat(platform->trackedNCLvars.perDeviceServiceRoot, platform->trackedNCLvars.uniquenessSuffix);
	strcat(platform->trackedNCLvars.perDeviceServiceRoot, ")");
    }

    platform->trackedNCLvars.lenPerDeviceServiceRoot = strlen(platform->trackedNCLvars.perDeviceServiceRoot);

    if (XRegionPSRunning() == PS_UP) {
	nuint8 servicesFlag = 0;
	bool temp;
	if (OPBSetPSDeviceOnEnabled("LPR")) servicesFlag |= LPR_SERVICE_MASK;
	if (OPBSetPSDeviceOnEnabled("AppSocket")) servicesFlag |= PORT9100_SERVICE_MASK;
	if (OPBSetPSDeviceOnEnabled("FTP")) servicesFlag |= FTP_SERVICE_MASK;
	if (OPBSetPSDeviceOnEnabled("IPP")) servicesFlag |= IPP_SERVICE_MASK;
	if ((xpmGetParamBoolVal(XpmHandle_None, "HTTP.On", &temp, ParamAccLvl_System) == XpmStatus_OK) && temp) servicesFlag |= HTTP_SERVICE_MASK;
	//rip	if (OPBSetPSDeviceOnEnabled("MIME")) servicesFlag |= RIP_SERVICE_MASK;
	//telnet	if (OPBSetPSDeviceOnEnabled("LPR")) servicesFlag |= LPR_SERVICE_MASK;
	if (platform->trackedNCLvars.servicesFlag != servicesFlag) {
	    changed = TRUE;
	    platform->trackedNCLvars.servicesFlag = servicesFlag;
	}
    }
#ifdef MDNS_DEBUG
    else {
	lprintf("!!PS not up....faking every service\n");
	platform->trackedNCLvars.servicesFlag =  LPR_SERVICE_MASK |PORT9100_SERVICE_MASK|FTP_SERVICE_MASK|IPP_SERVICE_MASK|HTTP_SERVICE_MASK/*|TELNET_SERVICE_MASK*/;
    }
#endif /* MDNS_DEBUG */
    if ((xpmGetParamIntVal(XpmHandle_None, "LPR.DataPortNumber", &port, ParamAccLvl_System)) != XpmStatus_OK)
	port = 515;
    if (platform->trackedNCLvars.lprServicePort != port) {
	changed = TRUE;
	platform->trackedNCLvars.lprServicePort = port;
    }
    if ((xpmGetParamIntVal(XpmHandle_None, "AppSocket.DataPortNumber", &port, ParamAccLvl_System)) != XpmStatus_OK)
	port = 9100;
    if (platform->trackedNCLvars.port9100ServicePort != port) {
	changed = TRUE;
	platform->trackedNCLvars.port9100ServicePort = port;
    }

    /* TXT record information */
    // protocol-specific part of TXT records
    // prime the strings
    platform->trackedNCLvars.lprTxt[0] = '\0';
    platform->trackedNCLvars.ippTxt[0] = '\0';
    platform->trackedNCLvars.port9100Txt[0] = '\0';
    platform->trackedNCLvars.ftpTxt[0] = '\0';
    platform->trackedNCLvars.httpTxt[0] = '\0';
    platform->trackedNCLvars.commonTxt[0] = '\0';
    // add in the adminurl item
#if 0
    inet_ntoa_b(platform->trackedNCLvars.ipAddr, scratch);
#else
    strcpy (scratch, platform->trackedNCLvars.hostName);
    strcat (scratch, ".local.");
#endif
    opbAddPascalString(platform->trackedNCLvars.lprTxt, "adminurl=http://", scratch, NULL);
    opbAddPascalString(platform->trackedNCLvars.ippTxt, "adminurl=http://", scratch, NULL);
    opbAddPascalString(platform->trackedNCLvars.port9100Txt, "adminurl=http://", scratch, NULL);
    opbAddPascalString(platform->trackedNCLvars.ftpTxt, "adminurl=http://", scratch, NULL);
    // no adminurl needed for _http._tcp

    // additional per protocol printer flags
    // number of queues per protocol
    opbAddPascalString(platform->trackedNCLvars.lprTxt, "qtotal=1", NULL, NULL);
    opbAddPascalString(platform->trackedNCLvars.ippTxt, "qtotal=1", NULL, NULL);
    opbAddPascalString(platform->trackedNCLvars.port9100Txt, "qtotal=1", NULL, NULL);
    opbAddPascalString(platform->trackedNCLvars.ftpTxt, "qtotal=1", NULL, NULL);
    //ALERT: how can/do we need to allow the user to reset the protocol priority ?
    opbAddPascalString(platform->trackedNCLvars.lprTxt, "rp=auto", NULL, NULL);
    opbAddPascalString(platform->trackedNCLvars.lprTxt, "priority=30", NULL, NULL);
    opbAddPascalString(platform->trackedNCLvars.ippTxt, "rp=ipp", NULL, NULL);
    opbAddPascalString(platform->trackedNCLvars.ippTxt, "priority=60", NULL, NULL);
    opbAddPascalString(platform->trackedNCLvars.ippTxt, "pdl=application/postscript", NULL, NULL);
#if 0
    // ALERT: example to advertise multi pdls
    opbAddPascalString(platform->trackedNCLvars.port9100Txt, "pdl=application/postscript,application/pdf,application/vnd.hp-PCL", NULL, NULL);
#endif
    opbAddPascalString(platform->trackedNCLvars.port9100Txt, "priority=10", NULL, NULL);
    opbAddPascalString(platform->trackedNCLvars.ftpTxt, "priority=80", NULL, NULL);

    // common part of TXT records
    opbAddPascalString(platform->trackedNCLvars.commonTxt, "txtvers=1", NULL, NULL);
    if (xpmGetParamStrVal(XpmHandle_None, "internal.PrinterLocation", printerLocation, ID_STR_LENGTH_XPM + 1, ParamAccLvl_System) == XpmStatus_OK) {
	if ((strcmp(platform->trackedNCLvars.printerLocation, printerLocation) != 0) || (strlen(platform->trackedNCLvars.printerLocation) != strlen(printerLocation))){
	    changed = TRUE;
	    strcpy(platform->trackedNCLvars.printerLocation, printerLocation);
	}
    }
    if (platform->trackedNCLvars.printerLocation[0] != '\0')
        opbAddPascalString(platform->trackedNCLvars.commonTxt, "note=", platform->trackedNCLvars.printerLocation, NULL);
    if (modelName[0] != '\0') {
        opbAddPascalString(platform->trackedNCLvars.commonTxt, "product=(", modelName, ")");
	opbAddPascalString(platform->trackedNCLvars.commonTxt, "ty=Xerox ", modelName, NULL);
    }
    // unsupported, uncheckable, or does not need checking
    opbAddPascalString(platform->trackedNCLvars.commonTxt, "Bind=F", NULL, NULL);
    opbAddPascalString(platform->trackedNCLvars.commonTxt, "Copies=T", NULL, NULL);
    opbAddPascalString(platform->trackedNCLvars.commonTxt, "Sort=F", NULL, NULL);
    opbAddPascalString(platform->trackedNCLvars.commonTxt, "Binary=T", NULL, NULL);
    opbAddPascalString(platform->trackedNCLvars.commonTxt, "Transparent=T", NULL, NULL);
    opbAddPascalString(platform->trackedNCLvars.commonTxt, "TBCP=F", NULL, NULL);
    //ALERT: these are checkable in PS by cross-region calls
    if (XRegionPSRunning() == PS_UP) {
	char truefalse[2];
	truefalse[1] ='\0';
	if (PSFunction.haveStaple())
	    truefalse[0] = 'T';
	else
	    truefalse[0] = 'F';
	opbAddPascalString(platform->trackedNCLvars.commonTxt, "Staple=", truefalse, NULL);
	if (PSFunction.colorCapable())
	    truefalse[0] = 'T';
	else
	    truefalse[0] = 'F';
	opbAddPascalString(platform->trackedNCLvars.commonTxt, "Color=", truefalse, NULL);
	if (PSFunction.haveDuplex())
	    truefalse[0] = 'T';
	else
	    truefalse[0] = 'F';
	opbAddPascalString(platform->trackedNCLvars.commonTxt, "Duplex=", truefalse, NULL);
	if (PSFunction.haveCollation())
	    truefalse[0] = 'T';
	else
	    truefalse[0] = 'F';
	opbAddPascalString(platform->trackedNCLvars.commonTxt, "Collate=", truefalse, NULL);
	if (PSFunction.haveCustomSizes())
	    truefalse[0] = 'T';
	else
	    truefalse[0] = 'F';
	opbAddPascalString(platform->trackedNCLvars.commonTxt, "PaperCustom=", truefalse, NULL);
	//ALERT: need to find the max paper size by name
	//	opbAddPascalString(platform->trackedNCLvars.commonTxt, "PaperMax=tabloid", NULL, NULL);
    }

    if (changed && (m && m->p && m->p->commandPipe != ERROR )) {
	SendCommand((mDNS *)m, kMDNSPipeCommandCodeUpdateIP);
    }
    return;
}

void
examineNvvarsNP( const char *name __attribute__((unused)), ParamStatus status __attribute__((unused)), void *info __attribute__((unused)) )
{
    (void)examineNvvars(0, NULL, NULL);
    return;
}

void
linkChangeNotified(mDNS *m, int state)
{
    if (m && m->p && m->p->commandPipe != ERROR ) {
	if (state == LINKSTATEUP) {
	    SendCommand((mDNS *)m, kMDNSPipeCommandCodeIFLinkChange);
	}
	//else {
	// if the link state is down, there's no sense of attempting to send
	// any resource record updates
	//}
    }
}
#endif /* (OEM_TARGET_OS == VXWORKS) */
#endif /* MDNS_OEM_COMPONENT_INCLUDED */
